#!/home/builder/ci_310/django_1643016687141/_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placeh/bin/python
from django.core import management

if __name__ == "__main__":
    management.execute_from_command_line()
